--    ______   ______   ______   ______   ______   ______   ______   ______
--   |  __  | |  __  | |  _   | |      | |  _   | |  _   | |      | |  __  |
--   | (_   | | /    | | |_)  | |  /\  | | |_)  | | |_)  | | |    | | |__  |
--   | __) 1| | \__ 3| | | \ 1| | /--\1| | |_) 3| | |_) 3| | |__ 1| | |__ 1|
--   |______| |______| |______| |______| |______| |______| |______| |______|
--        ______   ______   ______   ______   ______   ______   ______
--       |  _   | |  _   | |  _   | |   _  | |  __  | |  __  | | ___  |
--       | |_)  | | |_)  | | / \  | |   |  | | |__  | | /    | |  |   |
--       | |   3| | | \ 1| | \_/ 1| |  _| 8| | |__ 1| | \__ 3| |  |  1|
--       |______| |______| |______| |______| |______| |______| |______|
--
--                           Group 13 *
--                            Nicolas Baly
--                             Mark Kornfilt
--                              Francois Bonzon
--                               Alok Menghrajani

--  Hello,
--
--  This is the body of the package programmed for the 2000-2001 Ada course at
--  the EPFL.  We were four students to work on this project.  We were given a
--  dictionary  file containing about  200'000 words.  Our goal was to write a
--  program that would search the dictionary given a set of available letters,
--  something similar to the game of Scrabble.
--
--  Our first step towards solving this problem was to analyse the dictionary.
--  We were able to extract a huge number of information. Our most useful data
--  was the total number of words in the dictionary,  the range of lengths and
--  weights and the number of maximum letter occurrence for each possible word
--  length and word weight. This was done by Nicolas Baly.
--
--  We then  started to think  about an efficient  algorithm.  We came up with
--  several algorithms that seemed interesting,  but we opted for a bit compa-
--  raison scheme.  What we noticed  is that we could encode  all words of the
--  dictionary with  115 bits,  because the sum  of all maximum  letter occur-
--  rences is 115. Since our computers have got 64 bits registers we can write
--  an algorithm that has  the advantages to be pretty quick  and almost cons-
--  tant  (the complexity  mainly depends on  the dictionary size,  which is a
--  constant).
--
--  Here is the idea  Francois Bonzon and  Alok Menghrajani came up with.  115
--  bits are required for our algorithm. We encoded each possible letter comb-
--  ination in the available  128 bits  (two 64 bits variables).  We then used
--  two logic 'AND' to determine whether the dictionary word should be kept or
--  discarded. Here is an example that might make things clearer. To simplify,
--  we will encode with 8 bits only, representing the following letters:
--
--  "1st O - 2nd O - 1st M - 2nd M - 1st N - 2nd N - 1st S - 2nd S"
--
--  So if the input to search is "Somon", it will be encoded into:
--     0   (because there is a first O)
--     0   (because there is a second O)
--     0   (because there is a first M)
--     1   (because there is no second M)
--     0   (because there is a first N)
--     1   (because there is no second M)
--     0   (because there is a first S)
--     1   (because there is no second S)
--  Which gives us     -> 00010101
--
--  Now, let's try to match "Somon" with the following dictionary words:
--  "Mom", "Noon", "Moon"
--  Mom is encoded into   10110000  (notice that when encoding words from the
--  Noon is encoded into  11001100   dictionary, 0 and 1 have the inverse
--  Moon is encoded into  11101000   meaning than in the normal encoding).
--
--  We calculated the  "magic" operation using a table of Karnaugh.  If we had
--  used for both encodings the format: 0 = absence and 1 = presence. We would
--  have obtained the following table, with 0 meaning a successful match.
--
--       Dic
--   _______
--  |   |   |         Which gives  us the equation  (NOT (Dic)) OR Input.  But
--  | 0 | 1 |         since we can precalculate the  NOT (Dic) we modified the
--  |---+---|         encoding and  equation using De Morgan's laws.
--  | 0 | 0 | Input
--  |___|___|
--
--  Let's do the "magic" operation between the input and the dictionary words:
--  "Somon"            -> 00010101
--  "Mom"              -> 10110000
--  "Somon" AND "Mom"  -> 00010000
--  Since the result is not 0, the word fails and is discarded.
--
--  Second word:                           Third word:
--  "Somon"            -> 00010101         "Somon"            -> 00010101
--  "Noon"             -> 11001100         "Moon"             -> 11101000
--  "Somon" AND "Noon" -> 00000100         "Somon" AND "Moon" -> 00000000
--  This word also fails!                  Yes, "Moon" is a result!
--
--  You can see  our algorithm is  pretty simple,  and it's the thinking  that
--  conducted us the finding the "magic" operation that might be complicated.
--
--  Mark  Kornfilt  sorted the  dictionary  so that we could  quickly stop the
--  search.  The dictionary is obviously  sorted by length and weight,  that's
--  why our dictionary is twice the size of the original.
--
--  The coding and ASCII art was done by Francois Bonzon.
--
--  Here is how you should compile our code (with the profiler enabled):
--  gnatmake -O2 -pg scrabble_utilities.adb -largs
--                   /soft/lib/gcc-lib/sparc-sun-solaris2.6/2.7.2.1/gmon.o
--
--  Our dictionary file is available here in the CO rooms:
--  ~fbonzon/webster_gr13.txt

with Scrabble_Types, Ada.Text_IO;

package body Scrabble_Utilities is

   --  structure used to encode a set of letters, each bit representing
   --  the presence (0) or absence (1) of a letter (in the dictionary,
   --  0 = absence and 1 = presence); each word in the dictionary or
   --  in an input to search is encoded using two of those
   type Code_Type is mod 2 ** 64;

   --  indicates a bit number of Code_Type
   subtype Bit_Type is Natural range 0 .. 63;

   --  respective minimum and maximum for lengths and weights in webster
   subtype Length_Type is Positive range 2 .. 24;
   subtype Weight_Type is Positive range 2 .. 56;

   --  number of words in dictionary
   subtype Index_Type is Positive range 1 .. 234_884;

   --  an entry of our dictionary
   type Word_Type is record
      Length : Length_Type;
      Letters : String (1 .. Length_Type'Last);
      Code_1, Code_2 : Code_Type;
      Is_Result : Boolean;
   end record;

   --  Scrabble_Types.Small_Letters redefined for faster computations
   subtype Small_Letters is Natural range 0 .. 25;

   --  our pre-processed version of the webster dictionary, full path is
   --  ~fbonzon/webster_gr13.txt
   Dictionary_Name : constant String := "webster_gr13.txt";

   --  maximum occurrence of letters in a single word of the dictionary
   Max_Occurrence : constant array (Small_Letters) of Natural :=
     (0 =>  6,  1 =>  4,  2 =>  5,  3 =>  5,  4 =>  7,  5 =>  4,  6 =>  4,
      7 =>  4,  8 =>  6,  9 =>  2, 10 =>  4, 11 =>  5, 12 =>  4, 13 =>  6,
     14 =>  6, 15 =>  4, 16 =>  2, 17 =>  4, 18 =>  8, 19 =>  5, 20 =>  5,
     21 =>  3, 22 =>  3, 23 =>  2, 24 =>  4, 25 =>  3);  --  total 115

   --  number of letters encoded in Code_1, others are in Code_2
   --  We have wisely choosen these encoding in order to reduce
   --  the number of second case tests.
   In_Code_1 : constant array (Small_Letters) of Natural :=
     (0 =>  4,  1 =>  2,  2 =>  3,  3 =>  2,  4 =>  4,  5 =>  2,  6 =>  2,
      7 =>  2,  8 =>  4,  9 =>  1, 10 =>  2, 11 =>  3, 12 =>  2, 13 =>  4,
     14 =>  4, 15 =>  2, 16 =>  1, 17 =>  3, 18 =>  4, 19 =>  3, 20 =>  3,
     21 =>  2, 22 =>  1, 23 =>  1, 24 =>  2, 25 =>  1); --  total 64

   --  for each letter, when adding current occurrence of that letter,
   --  indicates the bit position where to insert it in Code_1
   Pos_Code_1 : constant array (Small_Letters) of Bit_Type :=
     (0 =>  0,  1 =>  4,  2 =>  6,  3 =>  9,  4 => 11,  5 => 15,  6 => 17,
      7 => 19,  8 => 21,  9 => 25, 10 => 26, 11 => 28, 12 => 31, 13 => 33,
     14 => 37, 15 => 41, 16 => 43, 17 => 44, 18 => 47, 19 => 51, 20 => 54,
     21 => 57, 22 => 59, 23 => 60, 24 => 61, 25 => 63);

   --  for each letter, when adding current occurrence of that letter,
   --  indicates the bit position where to insert it in Code_2
   Pos_Code_2 : constant array (Small_Letters) of Bit_Type :=
     (0 =>  9,  1 => 13,  2 => 14,  3 => 17,  4 => 18,  5 => 23,  6 => 25,
      7 => 27,  8 => 27,  9 => 32, 10 => 32, 11 => 33, 12 => 36, 13 => 36,
     14 => 38, 15 => 42, 16 => 45, 17 => 44, 18 => 44, 19 => 49, 20 => 51,
     21 => 54, 22 => 56, 23 => 58, 24 => 58, 25 => 61);

   --  number of words in the dictionary for all word lengths
   Words_Length : constant array (Length_Type) of Natural :=
     (2 =>   155,  3 =>  1351,  4 =>  5110,  5 =>  9987,  6 => 17477,
      7 => 23734,  8 => 29926,  9 => 32380, 10 => 30866, 11 => 26011,
     12 => 20460, 13 => 14938, 14 =>  9762, 15 =>  5924, 16 =>  3377,
     17 =>  1813, 18 =>   842, 19 =>   428, 20 =>   198, 21 =>    82,
     22 =>    41, 23 =>    17, 24 =>     5); --  total 234_884

   --  number of words in the dictionary for all word weights
   Words_Weight : constant array (Weight_Type) of Natural :=
     (2 =>    55,  3 =>   260,  4 =>   801,  5 =>  1609,  6 =>  3178,
      7 =>  4848,  8 =>  7352,  9 =>  9458, 10 => 11877, 11 => 14179,
     12 => 15865, 13 => 16893, 14 => 17449, 15 => 17236, 16 => 16723,
     17 => 15602, 18 => 13903, 19 => 12444, 20 => 10762, 21 =>  9029,
     22 =>  7565, 23 =>  6077, 24 =>  4983, 25 =>  3948, 26 =>  3084,
     27 =>  2379, 28 =>  1841, 29 =>  1400, 30 =>  1057, 31 =>   775,
     32 =>   589, 33 =>   439, 34 =>   344, 35 =>   245, 36 =>   154,
     37 =>   129, 38 =>   107, 39 =>    59, 40 =>    53, 41 =>    39,
     42 =>    26, 43 =>    20, 44 =>    18, 45 =>     8, 46 =>     7,
     47 =>     3, 48 =>     3, 49 =>     1, 50 =>     3, 52 =>     2,
     53 =>     2, 56 =>     1, others => 0); --  total 234_884

   --  maximum occurrence of letters in a single word of the dictionary,
   --  for all word lengths (for all words up to that length)
   --  (values for lengths > 21 are the same as in Max_Occurrence)
   Max_Length : constant array (2 .. 21, Small_Letters) of Natural :=
     (2 => (2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
            1, 1, 1, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1),
      3 => (2, 2, 1, 2, 2, 2, 2, 2, 2, 1, 1, 2, 2,
            2, 2, 2, 1, 2, 2, 2, 2, 1, 2, 1, 2, 1),
      4 => (2, 3, 2, 3, 3, 3, 3, 2, 3, 2, 2, 3, 2,
            2, 2, 2, 1, 2, 3, 2, 2, 2, 2, 1, 2, 3),
      5 => (3, 3, 3, 3, 3, 3, 3, 2, 3, 2, 3, 3, 3,
            3, 3, 3, 1, 3, 3, 3, 3, 2, 2, 2, 2, 3),
      6 => (3, 3, 3, 4, 4, 3, 3, 2, 3, 2, 4, 3, 3,
            3, 4, 3, 2, 3, 4, 3, 3, 2, 3, 2, 3, 3),
      7 => (4, 3, 3, 4, 4, 3, 3, 3, 3, 2, 4, 4, 3,
            4, 4, 3, 2, 3, 4, 3, 3, 2, 3, 2, 3, 3),
      8 => (4, 3, 4, 4, 4, 4, 4, 3, 4, 2, 4, 4, 4,
            4, 4, 3, 2, 4, 5, 4, 4, 2, 3, 2, 3, 3),
      9 => (5, 4, 4, 4, 5, 4, 4, 3, 4, 2, 4, 4, 4,
            5, 4, 3, 2, 4, 5, 4, 4, 2, 3, 2, 3, 3),
     10 => (5, 4, 4, 4, 5, 4, 4, 3, 4, 2, 4, 4, 4,
            5, 4, 4, 2, 4, 5, 4, 4, 2, 3, 2, 3, 3),
     11 => (5, 4, 4, 4, 5, 4, 4, 3, 4, 2, 4, 4, 4,
            5, 5, 4, 2, 4, 5, 4, 4, 3, 3, 2, 3, 3),
     12 => (5, 4, 5, 4, 5, 4, 4, 3, 5, 2, 4, 5, 4,
            5, 5, 4, 2, 4, 6, 4, 5, 3, 3, 2, 3, 3),
     13 => (5, 4, 5, 4, 5, 4, 4, 3, 5, 2, 4, 5, 4,
            5, 5, 4, 2, 4, 6, 4, 5, 3, 3, 2, 3, 3),
     14 => (5, 4, 5, 4, 6, 4, 4, 3, 6, 2, 4, 5, 4,
            5, 6, 4, 2, 4, 7, 5, 5, 3, 3, 2, 3, 3),
     15 => (5, 4, 5, 4, 6, 4, 4, 4, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 7, 5, 5, 3, 3, 2, 3, 3),
     16 => (5, 4, 5, 5, 6, 4, 4, 4, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 7, 5, 5, 3, 3, 2, 3, 3),
     17 => (5, 4, 5, 5, 6, 4, 4, 4, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 7, 5, 5, 3, 3, 2, 3, 3),
     18 => (6, 4, 5, 5, 6, 4, 4, 4, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 8, 5, 5, 3, 3, 2, 3, 3),
     19 => (6, 4, 5, 5, 6, 4, 4, 4, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 8, 5, 5, 3, 3, 2, 3, 3),
     20 => (6, 4, 5, 5, 6, 4, 4, 4, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 8, 5, 5, 3, 3, 2, 3, 3),
     21 => (6, 4, 5, 5, 6, 4, 4, 4, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 8, 5, 5, 3, 3, 2, 3, 3));

   --  maximum occurrence of letters in a single word of the dictionary,
   --  for all word weights (for all words up to that weight)
   --  (values for weights > 41 are the same as in Max_Occurrence)
   Max_Weight : constant array (2 .. 41, Small_Letters) of Natural :=
     (2 => (2, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 1, 0,
            1, 1, 0, 0, 1, 1, 1, 1, 0, 0, 0, 0, 0),
      3 => (2, 0, 0, 1, 2, 0, 1, 0, 2, 0, 0, 2, 0,
            2, 2, 0, 0, 2, 2, 2, 2, 0, 0, 0, 0, 0),
      4 => (2, 1, 1, 1, 2, 0, 1, 0, 2, 0, 0, 3, 1,
            2, 2, 1, 0, 2, 3, 2, 2, 0, 0, 0, 0, 0),
      5 => (3, 1, 1, 2, 3, 1, 2, 1, 2, 0, 0, 3, 1,
            3, 2, 1, 0, 3, 3, 3, 3, 1, 1, 0, 1, 0),
      6 => (3, 1, 1, 2, 4, 1, 2, 1, 3, 0, 1, 3, 1,
            3, 4, 1, 0, 3, 4, 3, 3, 1, 1, 0, 1, 0),
      7 => (3, 2, 1, 3, 4, 1, 3, 1, 3, 0, 1, 3, 2,
            4, 4, 2, 0, 3, 4, 3, 3, 1, 1, 0, 1, 0),
      8 => (4, 2, 2, 3, 4, 1, 3, 1, 3, 0, 1, 4, 2,
            4, 4, 2, 0, 4, 4, 3, 3, 1, 1, 0, 1, 0),
      9 => (4, 2, 2, 3, 4, 2, 3, 2, 3, 1, 1, 4, 2,
            5, 4, 2, 0, 4, 5, 4, 3, 1, 2, 1, 2, 0),
     10 => (4, 3, 2, 4, 4, 2, 3, 2, 4, 1, 1, 4, 2,
            5, 4, 2, 0, 4, 5, 4, 4, 2, 2, 1, 2, 0),
     11 => (5, 3, 3, 4, 4, 2, 3, 2, 4, 1, 1, 4, 3,
            5, 4, 3, 0, 4, 5, 4, 4, 2, 2, 1, 2, 1),
     12 => (5, 3, 3, 4, 5, 2, 4, 2, 4, 1, 2, 4, 3,
            5, 4, 3, 1, 4, 5, 4, 4, 2, 2, 1, 2, 1),
     13 => (5, 3, 3, 4, 5, 3, 4, 2, 4, 1, 2, 4, 3,
            5, 4, 3, 1, 4, 6, 4, 4, 2, 2, 1, 2, 1),
     14 => (5, 3, 3, 4, 5, 3, 4, 2, 4, 1, 2, 5, 3,
            5, 4, 3, 1, 4, 7, 4, 5, 2, 2, 1, 2, 1),
     15 => (5, 3, 3, 4, 5, 3, 4, 2, 4, 1, 2, 5, 3,
            5, 4, 3, 1, 4, 7, 4, 5, 2, 3, 1, 2, 1),
     16 => (5, 3, 3, 4, 5, 3, 4, 3, 5, 1, 2, 5, 3,
            5, 5, 3, 1, 4, 7, 4, 5, 2, 3, 1, 2, 1),
     17 => (5, 4, 4, 4, 6, 3, 4, 3, 5, 1, 3, 5, 4,
            5, 5, 3, 1, 4, 7, 5, 5, 2, 3, 1, 2, 1),
     18 => (5, 4, 4, 4, 6, 3, 4, 3, 5, 2, 3, 5, 4,
            5, 5, 4, 1, 4, 7, 5, 5, 2, 3, 1, 2, 1),
     19 => (5, 4, 4, 4, 6, 3, 4, 3, 5, 2, 3, 5, 4,
            6, 6, 4, 1, 4, 7, 5, 5, 2, 3, 2, 3, 1),
     20 => (5, 4, 4, 4, 6, 4, 4, 3, 5, 2, 3, 5, 4,
            6, 6, 4, 1, 4, 8, 5, 5, 2, 3, 2, 3, 1),
     21 => (5, 4, 4, 4, 6, 4, 4, 3, 5, 2, 3, 5, 4,
            6, 6, 4, 1, 4, 8, 5, 5, 2, 3, 2, 3, 1),
     22 => (5, 4, 4, 4, 6, 4, 4, 3, 5, 2, 4, 5, 4,
            6, 6, 4, 1, 4, 8, 5, 5, 3, 3, 2, 3, 2),
     23 => (6, 4, 4, 4, 6, 4, 4, 3, 6, 2, 4, 5, 4,
            6, 6, 4, 1, 4, 8, 5, 5, 3, 3, 2, 3, 2),
     24 => (6, 4, 4, 4, 6, 4, 4, 3, 6, 2, 4, 5, 4,
            6, 6, 4, 1, 4, 8, 5, 5, 3, 3, 2, 3, 2),
     25 => (6, 4, 4, 4, 6, 4, 4, 3, 6, 2, 4, 5, 4,
            6, 6, 4, 1, 4, 8, 5, 5, 3, 3, 2, 3, 2),
     26 => (6, 4, 4, 5, 6, 4, 4, 3, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 8, 5, 5, 3, 3, 2, 3, 2),
     27 => (6, 4, 5, 5, 6, 4, 4, 3, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 8, 5, 5, 3, 3, 2, 3, 2),
     28 => (6, 4, 5, 5, 6, 4, 4, 3, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 8, 5, 5, 3, 3, 2, 3, 2),
     29 => (6, 4, 5, 5, 6, 4, 4, 3, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 8, 5, 5, 3, 3, 2, 3, 2),
     30 => (6, 4, 5, 5, 6, 4, 4, 3, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 8, 5, 5, 3, 3, 2, 3, 2),
     31 => (6, 4, 5, 5, 7, 4, 4, 3, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 8, 5, 5, 3, 3, 2, 3, 3),
     32 => (6, 4, 5, 5, 7, 4, 4, 3, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 8, 5, 5, 3, 3, 2, 3, 3),
     33 => (6, 4, 5, 5, 7, 4, 4, 3, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 8, 5, 5, 3, 3, 2, 3, 3),
     34 => (6, 4, 5, 5, 7, 4, 4, 3, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 8, 5, 5, 3, 3, 2, 3, 3),
     35 => (6, 4, 5, 5, 7, 4, 4, 4, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 8, 5, 5, 3, 3, 2, 3, 3),
     36 => (6, 4, 5, 5, 7, 4, 4, 4, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 8, 5, 5, 3, 3, 2, 3, 3),
     37 => (6, 4, 5, 5, 7, 4, 4, 4, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 8, 5, 5, 3, 3, 2, 3, 3),
     38 => (6, 4, 5, 5, 7, 4, 4, 4, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 8, 5, 5, 3, 3, 2, 3, 3),
     39 => (6, 4, 5, 5, 7, 4, 4, 4, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 8, 5, 5, 3, 3, 2, 3, 3),
     40 => (6, 4, 5, 5, 7, 4, 4, 4, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 8, 5, 5, 3, 3, 2, 3, 3),
     41 => (6, 4, 5, 5, 7, 4, 4, 4, 6, 2, 4, 5, 4,
            6, 6, 4, 2, 4, 8, 5, 5, 3, 3, 2, 3, 3));

   --  pre-calculated powers of 2 inversed, used to clear a bit in a Code_Type
   --  (initialized in Load)
   Clear_Bit : array (Bit_Type) of Code_Type;

   --  the dictionaries, one sorted in decreasing length order,
   --  and the other sorted in decreasing weight order
   Dictionary_Length, Dictionary_Weight : array (Index_Type) of Word_Type;

   --  indexes to the first and last word for all word lengths
   Start_Length, End_Length : array (Length_Type) of Index_Type;

   --  indexes to the first and last word for all word weights
   Start_Weight, End_Weight : array (Weight_Type) of Index_Type;

   --  during search, set to True when at least one word is found
   Found_Result : Boolean := False;

   ------------
   --  Load  --
   ------------

   procedure Load is

      Dictionary_File : Ada.Text_IO.File_Type;
      Index : Positive;
      Letters : String (1 .. Length_Type'Succ (Length_Type'Last));
      Last : Length_Type;
      Letter : Small_Letters;
      Letters_Map : array (Small_Letters) of Natural;
      Code_1, Code_2 : Code_Type;

   begin --  Load

      --  initialization of variable Clear_Bit
      for I in Clear_Bit'Range loop
         Clear_Bit (I) := not (2 ** I);
      end loop;

      Ada.Text_IO.Put ("Loading dictionary.");
      Ada.Text_IO.Open (Dictionary_File, Ada.Text_IO.In_File, Dictionary_Name);
      --  create dictionary for Max_Length searches
      Ada.Text_IO.Skip_Line (Dictionary_File); --  first line has comments
      Index := Positive'First;
      for I in reverse Length_Type loop
         --  set starting point for I letters words
         Start_Length (I) := Index;
         for J in 1 .. Words_Length (I) loop
            Letters_Map := (others => 0);
            Code_1 := Code_Type'Last;
            Code_2 := Code_Type'Last;
            Ada.Text_IO.Get_Line (Dictionary_File, Letters, Last);
            --  first letter special case, might be uppercase
            if Letters (1) < Scrabble_Types.Small_Letters'First then
               Letter := Character'Pos (Letters (1)) - Character'Pos ('A');
            else
               Letter := Character'Pos (Letters (1)) -
                 Character'Pos (Scrabble_Types.Small_Letters'First);
            end if;
            Code_1 := Code_1 and
              Clear_Bit (Pos_Code_1 (Letter) + Letters_Map (Letter));
            Letters_Map (Letter) := Natural'Succ (Letters_Map (Letter));
            --  insert other letters...
            for K in 2 .. Last loop
               Letter := Character'Pos (Letters (K)) -
                 Character'Pos (Scrabble_Types.Small_Letters'First);
               if Letters_Map (Letter) < In_Code_1 (Letter) then
                  --  ...into first code
                  Code_1 := Code_1 and
                    Clear_Bit (Pos_Code_1 (Letter) + Letters_Map (Letter));
               else
                  --  ...into second code
                  Code_2 := Code_2 and
                    Clear_Bit (Pos_Code_2 (Letter) + Letters_Map (Letter));
               end if;
               Letters_Map (Letter) := Natural'Succ (Letters_Map (Letter));
            end loop;
            --  create dictionary entry
            Dictionary_Length (Index) :=
              (Length => Last,
               Letters => Letters (1 .. Length_Type'Last),
               Code_1 => not Code_1,
               Code_2 => not Code_2,
               Is_Result => False);
            Index := Positive'Succ (Index);
         end loop;
         --  set ending point for I letters words
         End_Length (I) := Positive'Pred (Index);
      end loop;
      Ada.Text_IO.Put (".");
      --  create dictionary for Max_Weight searches
      Ada.Text_IO.Skip_Line (Dictionary_File); --  line has comments
      Index := Positive'First;
      for I in reverse Weight_Type loop
         --  set starting point for with weight I
         Start_Weight (I) := Index;
         for J in 1 .. Words_Weight (I) loop
            Letters_Map := (others => 0);
            Code_1 := Code_Type'Last;
            Code_2 := Code_Type'Last;
            Ada.Text_IO.Get_Line (Dictionary_File, Letters, Last);
            --  first letter special case, might be uppercase
            if Letters (1) < Scrabble_Types.Small_Letters'First then
               Letter := Character'Pos (Letters (1)) - Character'Pos ('A');
            else
               Letter := Character'Pos (Letters (1)) -
                 Character'Pos (Scrabble_Types.Small_Letters'First);
            end if;
            Code_1 := Code_1 and
              Clear_Bit (Pos_Code_1 (Letter) + Letters_Map (Letter));
            Letters_Map (Letter) := Natural'Succ (Letters_Map (Letter));
            --  insert other letters...
            for K in 2 .. Last loop
               Letter := Character'Pos (Letters (K)) -
                 Character'Pos (Scrabble_Types.Small_Letters'First);
               if Letters_Map (Letter) < In_Code_1 (Letter) then
                  --  ... into first code
                  Code_1 := Code_1 and
                    Clear_Bit (Pos_Code_1 (Letter) + Letters_Map (Letter));
               else
                  --  ...into second code
                  Code_2 := Code_2 and
                    Clear_Bit (Pos_Code_2 (Letter) + Letters_Map (Letter));
               end if;
               Letters_Map (Letter) := Natural'Succ (Letters_Map (Letter));
            end loop;
            --  create dictionary entry
            Dictionary_Weight (Index) :=
              (Length => Last,
               Letters => Letters (1 .. Length_Type'Last),
               Code_1 => not Code_1,
               Code_2 => not Code_2,
               Is_Result => False);
            Index := Positive'Succ (Index);
         end loop;
         --  set ending point for with weight I
         End_Weight (I) := Positive'Pred (Index);
      end loop;
      Ada.Text_IO.Close (Dictionary_File);
      Ada.Text_IO.Put (". Done.");
      Ada.Text_IO.New_Line;

   end Load;

   -------------------------------
   --  Search_Words_Max_Length  --
   -------------------------------

   procedure Search_Words_Max_Length (Letters : in String) is

      --  tell compiler to optimize code for speed, see RM 2.8(23), needs
      --  option -O2 in the gnatmake compile command
      pragma Optimize (Time);

      --  tell compiler to suppress all checks for Constraint_Error,
      --  see RM 11.5(4)
      pragma Suppress (All_Checks);

      --  the word length where searching in the dictionary begins
      Start : Natural := Letters'Length;

      --  the encoded input
      Code_1, Code_2 : Code_Type := Code_Type'Last;

   begin --  Search_Words_Max_Length

      --  don't need to search anything if we don't have at least 2 letters
      if Start < Length_Type'First then
         return;
      end if;

      declare

         --  the occurrences of all letters found so far
         Letters_Map : array (Small_Letters) of Natural :=
           (others => Natural'First);

      begin

         --  encode the input
         declare
            --  the character beeing currently added
            Letter : Small_Letters;
         begin
            for I in Letters'Range loop
               if Letters (I) < Scrabble_Types.Small_Letters'First then
                  --  uppercase letter found
                  Letter := Character'Pos (Letters (I)) - Character'Pos ('A');
               else
                  --  lowercase letter found
                  Letter := Character'Pos (Letters (I)) -
                    Character'Pos (Scrabble_Types.Small_Letters'First);
               end if;
               if Letters_Map (Letter) < In_Code_1 (Letter) then
                  --  insert letter in first code
                  Code_1 := Code_1 and
                    Clear_Bit (Pos_Code_1 (Letter) + Letters_Map (Letter));
                  Letters_Map (Letter) := Natural'Succ (Letters_Map (Letter));
               elsif Letters_Map (Letter) < Max_Occurrence (Letter) then
                  --  insert letter in second code
                  Code_2 := Code_2 and
                    Clear_Bit (Pos_Code_2 (Letter) + Letters_Map (Letter));
                  Letters_Map (Letter) := Natural'Succ (Letters_Map (Letter));
               else
                  --  useless letter found (e.g. a seventh 'a')
                  Start := Length_Type'Pred (Start);
               end if;
            end loop;
         end;

         --  try to find even more useless letters, depending on the length
         --  (e.g. we can ignore a fifth 'a' when we have 8 letters in our
         --  set, because we know for all words with 8 or less letters, we
         --  have at most 4 'a's)
         if Start > Length_Type'Last then
            Start := Length_Type'Last;
         elsif Start <= 21 then
            --  (values for lengths > 21 are the same as in Max_Occurrence)
            declare
               --  value of Start in previous loop
               Prev : Natural;
            begin
               loop
                  Prev := Start;
                  for Letter in Small_Letters loop
                     if Letters_Map (Letter) > Max_Length (Prev, Letter) then
                        --  useless letter found
                        loop
                           Start := Length_Type'Pred (Start);
                           Letters_Map (Letter) :=
                             Natural'Pred (Letters_Map (Letter));
                           exit when Letters_Map (Letter) =
                             Max_Length (Prev, Letter);
                        end loop;
                        exit;
                     end if;
                  end loop;
                  --  we can stop when no useless letter found in last loop
                  exit when Prev = Start;
                  --  abort search if less than 2 letters are left
                  if Start < Length_Type'First then
                     return;
                  end if;
               end loop;
            end;
         end if;

      end;

      --  MAIN LOOP : search the dictionary
      for Length in reverse Length_Type'First .. Start loop
         for Index in Start_Length (Length) .. End_Length (Length) loop
            if (Code_1 and Dictionary_Length (Index).Code_1) =
              Code_Type'First and then (Code_2 and
              Dictionary_Length (Index).Code_2) = Code_Type'First then
               --  result found
               Dictionary_Length (Index).Is_Result := True;
               Found_Result := True;
            end if;
         end loop;
         exit when Found_Result;
      end loop;

   end Search_Words_Max_Length;

   -------------------------------
   --  Search_Words_Max_Weight  --
   -------------------------------

   procedure Search_Words_Max_Weight (Letters : in String) is

      --  tell compiler to optimize code for speed, see RM 2.8(23), needs
      --  option -O2 in the gnatmake compile command
      pragma Optimize (Time);

      --  tell compiler to suppress all checks for Constraint_Error,
      --  see RM 11.5(4)
      pragma Suppress (All_Checks);

      --  Scrabble_Types.Weights redefined with new indexes
      Weight : constant array (Small_Letters) of Positive :=
        (16 | 25 => 10, 9 | 23 => 8, 10 => 5, 5 | 7 | 21 | 22 | 24 => 4,
         1 | 2 | 12 | 15 => 3, 3 | 6 => 2, others => 1);

      --  the word weight where searching in the dictionary begins
      Start : Natural := Natural'First;

      --  the encoded input
      Code_1, Code_2 : Code_Type := Code_Type'Last;

   begin --  Search_Words_Max_Weight

      --  don't need to search anything if we don't have at least 2 letters
      if Letters'Length < Length_Type'First then
         return;
      end if;

      declare

         --  the occurrences of all letters found so far
         Letters_Map : array (Small_Letters) of Natural :=
           (others => Natural'First);

      begin

         --  encode the input
         declare
            --  the character beeing currently added
            Letter : Small_Letters;
         begin
            for I in Letters'Range loop
               if Letters (I) < Scrabble_Types.Small_Letters'First then
                  --  uppercase letter found
                  Letter := Character'Pos (Letters (I)) - Character'Pos ('A');
               else
                  --  lowercase letter found
                  Letter := Character'Pos (Letters (I)) -
                    Character'Pos (Scrabble_Types.Small_Letters'First);
               end if;
               if Letters_Map (Letter) < In_Code_1 (Letter) then
                  --  insert letter in first code
                  Code_1 := Code_1 and
                    Clear_Bit (Pos_Code_1 (Letter) + Letters_Map (Letter));
                  Letters_Map (Letter) := Natural'Succ (Letters_Map (Letter));
                  Start := Start + Weight (Letter);
               elsif Letters_Map (Letter) < Max_Occurrence (Letter) then
                  --  insert letter in second code
                  Code_2 := Code_2 and
                    Clear_Bit (Pos_Code_2 (Letter) + Letters_Map (Letter));
                  Letters_Map (Letter) := Natural'Succ (Letters_Map (Letter));
                  Start := Start + Weight (Letter);
               end if;
            end loop;
         end;

         --  try to find even more useless letters, depending on the weight
         --  (e.g. we can ignore a fifth 'a' when we have 10 points in our
         --  set, because we know for all words with 10 or less points, we
         --  have at most 4 'a's)
         if Start > Weight_Type'Last then
            Start := Weight_Type'Last;
         elsif Start <= 41 then
            --  (values for weights > 41 are the same as in Max_Occurrence)
            declare
               --  value of Start in previous loop
               Prev : Natural;
            begin
               loop
                  Prev := Start;
                  for Letter in Small_Letters loop
                     if Letters_Map (Letter) > Max_Weight (Prev, Letter) then
                        --  useless letter found
                        loop
                           Start := Start - Weight (Letter);
                           Letters_Map (Letter) :=
                             Natural'Pred (Letters_Map (Letter));
                           exit when Letters_Map (Letter) =
                             Max_Weight (Prev, Letter);
                        end loop;
                        exit;
                     end if;
                  end loop;
                  --  we can stop when no useless letter found in last loop
                  exit when Prev = Start;
                  --  abort search if less than 2 letters are left
                  if Start < Length_Type'First then
                     return;
                  end if;
               end loop;
            end;
         end if;

      end;

      --  MAIN LOOP : search the dictionary
      for Weight in reverse Weight_Type'First .. Start loop
         for Index in Start_Weight (Weight) .. End_Weight (Weight) loop
            if (Code_1 and Dictionary_Weight (Index).Code_1) =
              Code_Type'First and then (Code_2 and
              Dictionary_Weight (Index).Code_2) = Code_Type'First then
               --  result found
               Dictionary_Weight (Index).Is_Result := True;
               Found_Result := True;
            end if;
         end loop;
         exit when Found_Result;
      end loop;

   end Search_Words_Max_Weight;

   ---------------------
   --  Print_Results  --
   ---------------------

   --  This procedure is not optimized! Both length-sorted and weight-sorted
   --  dictionaries are browsed in entire, looking for words part of the
   --  result. We don't really care, cause time spent in this procedure is
   --  not measured.

   procedure Print_Results is
   begin --  Print_Results

      if Found_Result then
         --  browse Dictionary_Length
         for Index in Index_Type loop
            if Dictionary_Length (Index).Is_Result then
               Ada.Text_IO.Put_Line (Dictionary_Length (Index).Letters
                                     (1 .. Dictionary_Length (Index).Length));
            end if;
         end loop;
         --  browse Dictionary_Weight
         for Index in Index_Type loop
            if Dictionary_Weight (Index).Is_Result then
               Ada.Text_IO.Put_Line (Dictionary_Weight (Index).Letters
                                     (1 .. Dictionary_Weight (Index).Length));
            end if;
         end loop;
      else
         Ada.Text_IO.Put_Line ("Nothing found.");
      end if;

   end Print_Results;

   --------------------------------
   --  Save_And_CleanUp_Results  --
   --------------------------------

   --  This procedure is not optimized! Both length-sorted and weight-sorted
   --  dictionaries are browsed in entire, looking for words part of the
   --  result. We don't really care, cause time spent in this procedure is
   --  not measured.

   procedure Save_And_CleanUp_Results (File_Name : in String) is

      Results_File : Ada.Text_IO.File_Type;

   begin --  Save_And_CleanUp_Results

      Ada.Text_IO.Create (Results_File, Ada.Text_IO.Out_File, File_Name);

      if Found_Result then
         --  browse Dictionary_Length
         for Index in Index_Type loop
            if Dictionary_Length (Index).Is_Result then
               --  save result
               Ada.Text_IO.Put_Line (Results_File,
                                     Dictionary_Length (Index).Letters
                                     (1 .. Dictionary_Length (Index).Length));
               --  clean up result
               Dictionary_Length (Index).Is_Result := False;
            end if;
         end loop;
         --  browse Dictionary_Weight
         for Index in Index_Type loop
            if Dictionary_Weight (Index).Is_Result then
               --  save result
               Ada.Text_IO.Put_Line (Results_File,
                                     Dictionary_Weight (Index).Letters
                                     (1 .. Dictionary_Weight (Index).Length));
               --  clean up result
               Dictionary_Weight (Index).Is_Result := False;
            end if;
         end loop;
         Found_Result := False;
      end if;

      Ada.Text_IO.Close (Results_File);

   end Save_And_CleanUp_Results;

   ----------------
   --  Clean_Up  --
   ----------------

   --  We don't do anything because we don't have any pointers.

   procedure Clean_Up is
   begin --  Clean_Up

      null;

   end Clean_Up;

end Scrabble_Utilities;
